const { logger: log } = require('@nuskin/logger')
const { DynamoDBClient } = require('@aws-sdk/client-dynamodb')
const {
    DynamoDBDocument,
    GetCommand,
    PutCommand,
    UpdateCommand,
    TransactWriteCommand,
    DeleteCommand,
    ScanCommand,
    BatchGetCommand,
    BatchWriteCommand,
    QueryCommand
} = require('@aws-sdk/lib-dynamodb')

const ddbClient = new DynamoDBClient()
const ddbDoc = DynamoDBDocument.from(ddbClient)
require('./globals');
const {FULFILLED} = require('./constants');

/**
 * Performs a batchGet operation on a single table
 * @param {Object[]} keys - array of keys (pKey|sKey) values
 * @param {String} tableName The DynamoDB table name
 * @param {Number} retryCount - optional
 * @param {Number} waitTime - optional
 * @return {Promise<any>}
 */
async function batchGet(keys, tableName = process.env.TABLE, retryCount = 0, waitTime = 100) {
    let getPromises = [];
    let unprocessed = [];
    let results = [];
    const [numBatches, batchSize] = _calculateBatchSize(keys.length, maxBatchGetSize); // eslint-disable-line no-undef

    for (let i = 0; i < numBatches; i++) {
        const params = {
            RequestItems: {
                [tableName]: {
                    Keys: keys.slice(i * batchSize, (i + 1) * batchSize)
                }
            }
        }
        getPromises.push(ddbDoc.send(new BatchGetCommand(params)));
    }

    [results, unprocessed] = await checkForResults(getPromises, results, unprocessed, tableName);

    if (unprocessed.length) {
        if (retryCount >= batchRetryCount) { // eslint-disable-line no-undef
            log.error({dynamoDbUtilsBatchGetUnprocessed: unprocessed});
        } else {
            await wait(waitTime);
            results = results.concat(await batchGet(unprocessed, tableName, retryCount + 1, waitTime + 100))
        }
    }

    return {
        Responses: {
            [tableName]: results
        }
    };
}

/**
 * A helper for batchGet. Checks to see if it is time to get results and
 *  if so gets and returns them in an array.
 *
 * @param {array} getPromises
 * @param {array} results
 * @param {array} unprocessed
 * @param {string} tableName
 * @return {Promise<*[]>}
 */
async function checkForResults(getPromises, results, unprocessed, tableName) {
    const batches = await Promise.allSettled(getPromises);
    // const batches = await Promise.all(getPromises);

    for (const batch of batches) {
        if (batch.status === FULFILLED) {
            results = results.concat(batch.value.Responses[tableName]);
            if (batch.value.UnprocessedKeys[tableName]) {
                unprocessed = unprocessed.concat(batch.value.UnprocessedKeys[tableName].Keys);
            }
        } else {
            log.error({batchGetBatchError: batch});
        }
    }
    return [results, unprocessed];
}

/**
 * Calculate the batch size given a maximum size per batch:
 * e.g., if numOfKeys = 450 and maxBatchSize = 100, it will return 90.
 * @param {number} numOfKeys The number of items we want to process
 * @param {number} maxBatchSize The most items we can process in one request
 * @return {number} The number of items to process per request such that
 * the requests have a similar number of items
 * @private
 */
const _calculateBatchSize = (numOfKeys, maxBatchSize) => {
    const batches = Math.floor(numOfKeys / maxBatchSize) +
        (numOfKeys % maxBatchSize > 0 ? 1 : 0); // handle odd number of items
    return [batches, Math.ceil(numOfKeys / batches)];
};

const getWriteBatches = (records, size) => {
    const batches = [];

    records.reduce((arr, rec) => {
        if (arr.length === 0) {
            batches.push(arr);
        }
        arr.push(rec);
        if (arr.length === size) {
            arr = [];
        }
        return arr;
    }, []);

    return batches;
}

/**
 * batchWrites a list of records to dynamo
 *
 * @param {object} records - a list of records to batch write
 * @param {string} tableName - the table to write the records to
 * @param {number} batchSize - number of records the async write before waiting
 * @param {number} retryCount - number of retries for this batch
 * @param {number} waitTime - wait time to proceed with unprocessed items.
 * @return {Promise<Qualification>}
 */
async function batchWrite(records, tableName=process.env.TABLE, batchSize = 1000, retryCount = 0, waitTime = 100) {
    const batches = getWriteBatches(records, batchSize);

    for (const recBatch of batches) {
        let putRequests = [];
        const promises = [];

        for (const record of recBatch) {
            putRequests.push({'PutRequest': {'Item': _copyData(record)}});

            if (putRequests.length === maxBatchWriteSize) { // eslint-disable-line no-undef
                promises.push(ddbDoc.send(new BatchWriteCommand({RequestItems: {[tableName]: putRequests}})));
                putRequests = [];
            }
        }

        if (putRequests.length > 0) {
            promises.push(ddbDoc.send(new BatchWriteCommand({RequestItems: {[tableName]: putRequests}})));
        }

        const results = await Promise.allSettled(promises);
        log.debug({dynamoUtilBatchWriteResponse: results});

        const unprocessed = [];
        for (const batch of results) {
            if (batch.status === FULFILLED) {
                if (batch.value.UnprocessedItems[tableName]) {
                    for (const unprocessedItem of batch.value.UnprocessedItems[tableName]) {
                        unprocessed.push(unprocessedItem.PutRequest.Item);
                    }
                }
            } else {
                log.error({batchWriteBatchError: batch});
            }
        }

        if (unprocessed.length > 0) {
            if (retryCount > batchRetryCount) { // eslint-disable-line no-undef
                log.error({dynamoUtilsBatchWriteUnprocessed: unprocessed});
            } else {
                await wait(waitTime);
                await batchWrite(unprocessed, tableName, batchSize, retryCount + 1, waitTime + 100);
            }
        }
    }
}

/**
 * Call DocumentClient.get with the given tableName (or default)
 * and given params.
 * @param {Object} params
 * @param {String} tableName
 * @return {Promise<Request<DocumentClient.GetItemOutput, AWSError>>}
 */
const get = (params, tableName=process.env.TABLE) => {
    return ddbDoc.send(new GetCommand(_copyData(params, tableName)));
};

/**
 * Call DocumentClient.query with the given tableName (or default)
 * and given params.
 * @param {Object} params
 * @param {String} tableName
 * @return {Promise<Request<DocumentClient.ScanItemOutput, AWSError>>}
 */
const query = (params, tableName=process.env.TABLE) => {
    return ddbDoc.send(new QueryCommand(_copyData(params, tableName)));
};

/**
 * Call DocumentClient.query with the given tableName (or default)
 * and given params.
 * @param {Object} params
 * @param {String} tableName
 * @return {Promise<Request<DocumentClient.ScanItemOutput, AWSError>>}
 */
const scan = (params, tableName=process.env.TABLE) => {
    return ddbDoc.send(new ScanCommand(_copyData(params, tableName)));
};

/**
 * Call DocumentClient.put with the given tableName (or default)
 * and given params.
 * @param {Object} params
 * @param {String} tableName
 * @return {Promise<Request<DocumentClient.PutItemOutput, AWSError>>}
 */
const put = (params, tableName=process.env.TABLE) => {
    return ddbDoc.send(new PutCommand(_copyData(params, tableName)));
};

/**
 * Call DocumentClient.update with the given tableName (or default)
 * and given params.
 * @param {Object} params
 * @param {String} tableName
 * @return {Promise<Request<DocumentClient.UpdateItemOutput, AWSError>>}
 */
const update = (params, tableName=process.env.TABLE) => {
    return ddbDoc.send(new UpdateCommand(_copyData(params, tableName)));
};

/**
 * Performs a transactWrite operation
 * @param {*} params - request parameters
 * @return {Promise<Object>}
 */
const transactWrite = (params) => {
    return ddbDoc.send(new TransactWriteCommand(params));
};

/**
 * Call DocumentClient.delete with the given tableName (or default)
 * and given params.
 * @param {Object} params
 * @param {String} tableName
 * @return {Promise<Request<DocumentClient.DeleteItemOutput, AWSError>>}
 */
const _delete = (params, tableName = process.env.TABLE) => {
    return ddbDoc.send(new DeleteCommand(_copyData(params, tableName)))
};

/**
 * Cleans an object the params object for DynamoDb calls
 *
 * @param {Object} params
 * @param {String} tableName
 * @return {Object}
 * @private
 */
const _copyData = (params, tableName = null) => {
    const retParams = JSON.parse(JSON.stringify(params));

    if (tableName) {
        retParams.TableName = tableName;
    }

    return retParams;
};

/**
 * Creates a promise that will be resolved in the specified amount of
 *  milliseconds. The calling function can call it with an await to delay
 *  execution of the code.
 *
 * @param {number} delay
 * @return {Promise<any>}
 */
const wait = (delay) => new Promise((resolve) => setTimeout(resolve, delay));

module.exports = {
    batchGet,
    batchWrite,
    get,
    query,
    scan,
    put,
    update,
    transactWrite,
    delete: _delete
};
